#include "phone.h"
#include "utils.h"
#include <speex/speex.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <alsa/asoundlib.h>

static char *audio_device;
static int audio_mode;     /* see main.c for values and meaning */

static SpeexBits bits;
static void *dec_state;

static SpeexBits encbits;
static void *enc_state;

/* 25 for a buffer of 500ms, each block being 20ms */

#define N 25
static int16_t cbuf[320*N];
static volatile int size;
static int read_head, write_head;
static lockwait buflock;

static void put_audio(int16_t *b)
{
  lw_lock(&buflock);

  if (size == N) {
    /* full buffer, kill oldest block */
    memcpy(cbuf + write_head, b, 320*2);
    read_head += 320; if (read_head == 320*N) read_head = 0;
    write_head += 320; if (write_head == 320*N) write_head = 0;
  } else {
    size++;
    memcpy(cbuf + write_head, b, 320*2);
    write_head += 320; if (write_head == 320*N) write_head = 0;
  }

  lw_unlock(&buflock);
}

static void get_audio(int16_t *b)
{
  while (size == 0) usleep(20 * 1000);
  lw_lock(&buflock);
  memcpy(b, cbuf + read_head, 320*2);
  read_head += 320; if (read_head == 320*N) read_head = 0;
  size--;
  lw_unlock(&buflock);
}

static volatile int run;
static lockwait runlock;

#define A do { if (err < 0) { fprintf(stderr, "%s:%d: alsa error '%s'\n", __FILE__, __LINE__, snd_strerror(err)); abort(); } } while (0)

static void xrun(snd_pcm_t *h)
{
  snd_pcm_prepare(h);
}

static void suspend(snd_pcm_t *h)
{
  int res;
  int loop = 0;
  while ((res = snd_pcm_resume(h)) == -EAGAIN && loop < 100) {
    struct timeval t;
    t.tv_sec = 0;
    t.tv_usec = 10 * 1000;
    select(0, 0, 0, 0, &t);
    loop++;
  }
  if (res < 0) snd_pcm_prepare(h);
}

static void alsa_play(snd_pcm_t *h, void *b)
{
#define X do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); xrun(h); } while (0)
#define Y do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); suspend(h); } while (0)
  size_t count = 320;
  ssize_t r;
  unsigned char *bb;

  bb = b;

  while (count) {
    r = snd_pcm_writei(h, bb, count);
    if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
      /* nothing */
    }
    else if (r == -EPIPE) { usleep(50*1000); X; }
    else if (r == -ESTRPIPE) Y;
    else if (r < 0) abort();
    if (r > 0) {
      count -= r;
      bb += r * 2;
    }
  }
#undef X
#undef Y
}

static int16_t b48[320*3*2];

static void alsa_record(snd_pcm_t *h, void *b)
{
#define X do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); xrun(h); } while (0)
#define Y do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); suspend(h); } while (0)
  size_t count = 320;
  ssize_t r;
  unsigned char *bb;
  int channels = 1;

  bb = b;

  if (audio_mode == 1) {
    count = 320*3;
    bb = (unsigned char *)b48;
    channels = 2;
  }

  while (count) {
    r = snd_pcm_readi(h, bb, count);
    if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
      /* nothing */
    }
    else if (r == -EPIPE) X;
    else if (r == -ESTRPIPE) Y;
    else if (r < 0) abort();
    if (r > 0) {
      count -= r;
      bb += r * 2 * channels;
    }
  }

  if (audio_mode == 1) {
    int16_t *b16 = b;
    int i;
    for (i = 0; i < 320; i++) {
      long v = (b48[i*3*2] + b48[i*3*2+2] + b48[i*3*2+4]) / 3;
      b16[i] = v;
    }
  }

#undef X
#undef Y
}

static void *audio_thread(void *_)
{
  snd_pcm_t           *h;
  snd_pcm_hw_params_t *hw_params;
  snd_pcm_sw_params_t *sw_params;
  snd_pcm_uframes_t   period_frames;
  snd_pcm_uframes_t   buffer_frames;
  unsigned int        rate;
  snd_pcm_format_t    format;
  int err;
  int16_t buf[320];

start:
  lw_lock(&runlock);
  while (!run) lw_wait(&runlock);
  lw_unlock(&runlock);

  lw_lock(&buflock);
  size = 0;
  read_head = write_head = 0;
  lw_unlock(&buflock);

  format = SND_PCM_FORMAT_S16_LE;
  err = snd_pcm_open(&h, "default", SND_PCM_STREAM_PLAYBACK, 0); A;
  snd_pcm_hw_params_alloca(&hw_params);
  snd_pcm_sw_params_alloca(&sw_params);
  err = snd_pcm_hw_params_any(h, hw_params); A;
  err = snd_pcm_hw_params_set_access(h, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED); A;
  err = snd_pcm_hw_params_set_format(h, hw_params, format); A;
  err = snd_pcm_hw_params_set_channels(h, hw_params, 1); A;
  rate = 16000;
  err = snd_pcm_hw_params_set_rate_near(h, hw_params, &rate, 0); A;
  if (rate != 16000) {printf("bad rate\n"); abort();}
  period_frames = 320;
  buffer_frames = 320*25;
  err = snd_pcm_hw_params_set_period_size_near(h,hw_params,&period_frames,0);A;
  err = snd_pcm_hw_params_set_buffer_size_near(h, hw_params, &buffer_frames);A;
  err = snd_pcm_hw_params(h, hw_params); A;

  while (run) {
    get_audio(buf);
    alsa_play(h, buf);
  }

  snd_pcm_close(h);

  goto start;

  return 0;
}

void audio_create(char *_audio_device, int _audio_mode)
{
  int i;
  audio_device = _audio_device;
  audio_mode = _audio_mode;
  speex_bits_init(&bits);
  dec_state = speex_decoder_init(&speex_wb_mode);
  speex_bits_init(&encbits);
  enc_state = speex_encoder_init(&speex_wb_mode);
  i = 1;
  speex_encoder_ctl(enc_state, SPEEX_SET_VBR, &i);
  i = 0;
  speex_encoder_ctl(enc_state, SPEEX_SET_VBR_QUALITY, &i);
  lw_init(&buflock);
  lw_init(&runlock);
  new_thread(audio_thread, NULL);
}

void audio_start(void)
{
  run = 1;
  lw_wakeup(&runlock);
}

void audio_stop(void)
{
  run = 0;
}

static int16_t decbuf[320];
void audio_speex_play(char *b, int l)
{
  speex_bits_read_from(&bits, b, l); 
  if (speex_decode_int(dec_state, &bits, decbuf))
    { printf("speex dec err\n"); return; }
  put_audio(decbuf);
}

static snd_pcm_t *hrec;

static int16_t encbuf[320];
int audio_speex_record(char *b, int l)
{
  alsa_record(hrec, encbuf);
  speex_bits_reset(&encbits);
  speex_encode_int(enc_state, encbuf, &encbits);
  return speex_bits_write(&encbits, b, l);
}

void audio_open_record(void)
{
  snd_pcm_hw_params_t *hw_params;
  snd_pcm_sw_params_t *sw_params;
  snd_pcm_uframes_t   period_frames;
  snd_pcm_uframes_t   buffer_frames;
  unsigned int        rate;
  snd_pcm_format_t    format;
  int err;
  unsigned int        refrate = 16000;
  snd_pcm_uframes_t   refframes = 320;
  int                 channels = 1;

  if (audio_mode == 1) {
    refrate = 48000;
    channels = 2;
    refframes = 320*3;
  }

  format = SND_PCM_FORMAT_S16_LE;
  err = snd_pcm_open(&hrec, audio_device, SND_PCM_STREAM_CAPTURE, 0); A;
  snd_pcm_hw_params_alloca(&hw_params);
  snd_pcm_sw_params_alloca(&sw_params);
  err = snd_pcm_hw_params_any(hrec, hw_params); A;
  err = snd_pcm_hw_params_set_access(hrec, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED); A;
  err = snd_pcm_hw_params_set_format(hrec, hw_params, format); A;
  err = snd_pcm_hw_params_set_channels(hrec, hw_params, channels); A;
  rate = refrate;
  err = snd_pcm_hw_params_set_rate_near(hrec, hw_params, &rate, 0); A;
  if (rate != refrate) {printf("bad rate\n"); abort();}
  period_frames = refframes;
  buffer_frames = refframes*25;
  err = snd_pcm_hw_params_set_period_size_near(hrec,hw_params,&period_frames,0);A;
  err = snd_pcm_hw_params_set_buffer_size_near(hrec, hw_params, &buffer_frames);A;
  err = snd_pcm_hw_params(hrec, hw_params); A;
}

void audio_close_record(void)
{
  snd_pcm_close(hrec);
}
